# Generated by Django 3.2.16 on 2022-12-13 12:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mozilla_django_oidc_db", "0010_openidconnectconfig_userinfo_claims_source"),
    ]

    operations = [
        migrations.AlterField(
            model_name="openidconnectconfig",
            name="userinfo_claims_source",
            field=models.CharField(
                choices=[
                    ("userinfo_endpoint", "Userinfo endpoint"),
                    ("id_token", "ID token"),
                ],
                default="userinfo_endpoint",
                help_text="Indicates the source from which the user information claims should be extracted.",
                max_length=100,
                verbose_name="user information claims extracted from",
            ),
        ),
    ]
